"""Lifetime regression

Notes
-----
This module contains two parametric lifetime regressions.
ProportionalHazard is not Cox regression (Cox is semiparametric).
"""

from __future__ import annotations

from abc import ABC
from typing import Any, Callable, Literal, Self, final

import numpy as np
import numpydoc.docscrape as docscrape  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import NDArray
from scipy.optimize import Bounds
from typing_extensions import overload, override

from relife.base import OptimizerConfig, ParametricModel
from relife.typing import (
    AnyFloat,
    NumpyBool,
    NumpyFloat,
    Seed,
)

from ._base import (
    FittableParametricLifetimeModel,
    FrozenParametricLifetimeModel,
    LifetimeData,
    LifetimeLikelihood,
    document_args,
)
from ._distribution import (
    Gamma,
    LifetimeDistribution,
    get_distrib_params_bounds,
    init_distrib_params_from_lifetimes,
)

__all__: list[str] = [
    "ParametricAcceleratedFailureTime",
    "ParametricProportionalHazard",
]


def _broadcast_time_covar(
    time: AnyFloat, covar: AnyFloat
) -> tuple[NumpyFloat, NumpyFloat]:
    time = np.atleast_2d(np.asarray(time))  #  (m, n)
    covar = np.atleast_2d(np.asarray(covar))  #  (m, nb_coef)
    match (time.shape[0], covar.shape[0]):
        case (1, _):
            time = np.repeat(time, covar.shape[0], axis=0)
        case (_, 1):
            covar = np.repeat(covar, time.shape[0], axis=0)
        case (m1, m2) if m1 != m2:
            raise ValueError(
                f"""
                Incompatible time and covar. time has {m1} nb_assets but
                covar has {m2} nb_assets
                """
            )
        case _:
            pass
    return time, covar


def _broadcast_time_covar_shapes(
    time_shape: tuple[int, ...], covar_shape: tuple[int, ...]
) -> tuple[int, ...]:
    """
    time_shape : (), (n,) or (m, n)
    covar_shape : (), (nb_coef,) or (m, nb_coef)
    """
    match [time_shape, covar_shape]:
        case [(), ()] | [(), (_,)]:
            return ()
        case [(), (m, _)]:
            return m, 1
        case [(n,), ()] | [(n,), (_,)]:
            return (n,)
        case [(n,), (m, _)] | [(m, n), ()] | [(m, n), (_,)]:
            return m, n
        case [(mt, n), (mc, _)] if mt != mc:
            if mt != 1 and mc != 1:
                raise ValueError(
                    f"""
                    Invalid time and covar : time got {mt} nb assets but covar
                    got {mc} nb assets
                    """
                )
            return max(mt, mc), n
        case [(mt, n), (mc, _)] if mt == mc:
            return mt, n
        case _:
            raise ValueError(
                f"Invalid time or covar shape. Got {time_shape} and {covar_shape}"
            )


@final
class LinearCovarEffect(ParametricModel):
    """
    Covariates effect.

    Parameters
    ----------
    coefficients : tuple of float, default is (None,)
        Coefficients of the covariates effect.
    """

    def __init__(self, coefficients: tuple[float | None, ...] = (None,)):
        super().__init__(**{f"coef_{i + 1}": v for i, v in enumerate(coefficients)})

    @property
    def nb_coef(self) -> int:
        """
        Returns the number of coefficients.

        Returns
        -------
        out : int
        """
        return self.nb_params

    def g(self, covar: AnyFloat) -> NumpyFloat:
        """
        Returns the covariates effect.

        Parameters
        ----------
        covar : float or np.ndarray
            The covariate values

        Returns
        -------
        out : np.float64 or np.ndarray
            If `covar.shape` is `()`, `out` is `float`.
            If `covar.shape` is `(nb_coef,)`, `out.shape` is `()`.
            If `covar.shape` is `(m, nb_coef)`, `out.shape` is `(m, 1)`.
        """
        arr_covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        if arr_covar.ndim > 2:
            raise ValueError(
                f"""
                Invalid covar shape. Expected (nb_coef,) or (m, nb_coef) but
                got {arr_covar.shape}
                """
            )
        covar_nb_coef = arr_covar.size if arr_covar.ndim <= 1 else arr_covar.shape[-1]
        if covar_nb_coef != self.nb_coef:
            raise ValueError(
                f"""
                Invalid covar. Number of covar does not match number of
                coefficients. Got {self.nb_coef} nb_coef but covar shape is
                {arr_covar.shape}
                """
            )
        g = np.exp(np.sum(self.params * arr_covar, axis=-1, keepdims=True))  # (m, 1)
        if arr_covar.ndim <= 1:
            return np.float64(g.item())
        return g

    def jac_g(self, covar: AnyFloat) -> NumpyFloat:
        """
        Returns the jacobian of the covariates effect.

        Parameters
        ----------
        covar : float or np.ndarray
            The covariate values

        Returns
        -------
        out : np.ndarray
            If `covar.shape` is `()` or `(nb_coef,)`, `out.shape` is `(nb_coef,)`.
            If `covar.shape` is (m, nb_coef)`, `out.shape` is `(nb_coef, m, 1)`.
        """
        arr_covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        g = self.g(arr_covar)  # () or (m, 1)
        jac = arr_covar.T.reshape(self.nb_coef, -1, 1) * g  # (nb_coef, m, 1)
        if arr_covar.ndim <= 1:
            jac = jac.reshape(self.nb_coef)  # (nb_coef,) or (nb_coef, m, 1)
        return jac  # (nb_coef, m, 1)


_covar_docstring = [
    docscrape.Parameter(
        "covar",
        "float or np.ndarray",
        [
            "Covariates values.",
            "float can only be valid if the regression has one coefficients.",
            "Otherwise it must be a ndarray of shape `(nb_coef,)` or `(m, nb_coef)`.",
        ],
    ),
]


class ParametricLifetimeRegression(FittableParametricLifetimeModel[AnyFloat], ABC):
    """
    Base class for lifetime regression.
    """

    baseline: LifetimeDistribution
    covar_effect: LinearCovarEffect

    def __init__(
        self,
        baseline: LifetimeDistribution,
        coefficients: tuple[float | None, ...] = (None,),
    ):
        super().__init__()
        self.covar_effect = LinearCovarEffect(coefficients)
        self.baseline = baseline

    @property
    def coefficients(self) -> NDArray[np.float64]:
        """
        Returns the coefficients values.

        Returns
        -------
        out : ndarray
        """
        return self.covar_effect.params

    @property
    def nb_coef(self) -> int:
        """
        Returns the number of coefficients.

        Returns
        -------
        out : int
        """
        return self.covar_effect.nb_params

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def sf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return super().sf(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def isf(self, probability: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        cumulative_hazard_rate = -np.log(
            np.clip(probability, 0, 1 - np.finfo(float).resolution)
        )
        return self.ichf(cumulative_hazard_rate, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def cdf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return super().cdf(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def pdf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return super().pdf(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def ppf(self, probability: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return super().ppf(probability, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def mrl(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return super().mrl(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        covar: AnyFloat,
        *,
        deg: int = 10,
    ) -> NumpyFloat:
        return super().ls_integrate(func, a, b, covar, deg=deg)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def moment(self, n: int, covar: AnyFloat) -> NumpyFloat:
        return super().moment(n, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def mean(self, covar: AnyFloat) -> NumpyFloat:
        return super().mean(covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def var(self, covar: AnyFloat) -> NumpyFloat:
        return super().var(covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def median(self, covar: AnyFloat) -> NumpyFloat:
        return super().median(covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def jac_sf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        return -self.jac_chf(time, covar) * self.sf(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def jac_cdf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        return super().jac_cdf(time, covar)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def jac_pdf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        jac = self.jac_hf(time, covar) * self.sf(time, covar) + self.jac_sf(
            time, covar
        ) * self.hf(time, covar)
        return jac

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel, args_docstring=_covar_docstring
    )
    def rvs(
        self,
        size: int | tuple[int, int],
        covar: AnyFloat,
        *,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        return super().rvs(
            size,
            covar,
            return_event=return_event,
            return_entry=return_entry,
            seed=seed,
        )

    def freeze(self, covar: AnyFloat) -> FrozenParametricLifetimeModel[AnyFloat]:
        """
        Freeze regression covar.

        Parameters
        ----------
        covar : float or np.ndarray
            Covariates values. float can only be valid if the regression has one coefficients.
            Otherwise it must be a ndarray of shape `(nb_coef,)` or `(m, nb_coef)`.

        Returns
        -------
        out: frozen regression
            The same object but with `covar` stored as object data. Calling methods
            from the frozen regression does not need `covar`.
        """
        return FrozenParametricLifetimeModel(self, covar)

    @override
    def init_likelihood(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> LifetimeLikelihood[Self]:
        assert isinstance(args, np.ndarray)
        regression = type(self)(
            type(self.baseline)(), coefficients=(0.0,) * np.atleast_2d(args).shape[-1]
        )  # init new regression object with appropriate number of covar
        lifetime_data = LifetimeData(time, args, event, entry)
        x0 = kwargs.get(
            "x0", init_regression_params_from_lifetimes(regression, lifetime_data)
        )
        regression.params = x0
        config = OptimizerConfig(x0)
        config.scipy_minimize_options["bounds"] = kwargs.get(
            "bounds", get_regression_params_bounds(regression)
        )
        config.scipy_minimize_options["method"] = kwargs.get("method", "L-BFGS-B")
        config.covariance_method = kwargs.get(
            "covariance_method",
            "2point" if isinstance(regression.baseline, Gamma) else "cs",
        )
        optimizer = LifetimeLikelihood(regression, lifetime_data, config)
        return optimizer

    @override
    def fit(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> Self:
        assert isinstance(args, np.ndarray)
        self.covar_effect = LinearCovarEffect(
            (None,) * np.atleast_2d(np.asarray(args, dtype=np.float64)).shape[-1]
        )  # changes params structure depending on number of covar
        return super().fit(time, args, event, entry, **kwargs)


def init_regression_params_from_lifetimes(
    model: ParametricLifetimeRegression, data: LifetimeData
) -> NDArray[np.float64]:
    param0 = np.zeros_like(model.params, dtype=np.float64)
    param0[-model.baseline.params.size :] = init_distrib_params_from_lifetimes(
        model.baseline, data
    )
    return param0


def get_regression_params_bounds(model: ParametricLifetimeRegression) -> Bounds:
    lb = np.concatenate(
        (
            np.full(model.covar_effect.nb_params, -np.inf),
            get_distrib_params_bounds(
                model.baseline
            ).lb,  # baseline has _params_bounds according to typing
        )
    )
    ub = np.concatenate(
        (
            np.full(model.covar_effect.nb_params, np.inf),
            get_distrib_params_bounds(model.baseline).ub,
        )
    )
    return Bounds(lb, ub)


@final
class ParametricProportionalHazard(ParametricLifetimeRegression):
    r"""
    Proportional Hazard regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = g(\beta, x) H_0(t) = e^{\beta \cdot x} H_0(t)

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : _CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Sun, J. (2006). The statistical analysis of interval-censored failure
        time data (Vol. 3, No. 1). New York: springer.

    See Also
    --------
    regression.AFT : Accelerated Failure Time regression.

    """

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def hf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return self.covar_effect.g(covar) * self.baseline.hf(time)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def chf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return self.covar_effect.g(covar) * self.baseline.chf(time)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def ichf(self, cumulative_hazard_rate: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return self.baseline.ichf(cumulative_hazard_rate / self.covar_effect.g(covar))

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def dhf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return self.covar_effect.g(covar) * self.baseline.dhf(time)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def jac_hf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = _broadcast_time_covar_shapes(
            time.shape, covar.shape
        )  # (), (n,) or (m, n)
        time, covar = _broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar)  # (nb_coef, m, 1)

        baseline_hf = self.baseline.hf(time)  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf = self.baseline.jac_hf(time)  # (p, m, n)
        jac_g = np.repeat(
            jac_g, baseline_hf.shape[-1], axis=-1
        )  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_hf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_hf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        return jac

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def jac_chf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = _broadcast_time_covar_shapes(
            time.shape, covar.shape
        )  # (), (n,) or (m, n)
        time, covar = _broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar)  # (nb_coef, m, 1)
        baseline_chf = self.baseline.chf(time)  # (m, n)
        #  p == baseline.nb_params
        baseline_jac_chf = self.baseline.jac_chf(time)  # (p, m, n)
        jac_g = np.repeat(
            jac_g, baseline_chf.shape[-1], axis=-1
        )  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                baseline_chf * jac_g,  #  (nb_coef, m, n)
                g * baseline_jac_chf,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        return jac


@final
class ParametricAcceleratedFailureTime(ParametricLifetimeRegression):
    # noinspection PyUnresolvedReferences
    r"""
    Accelerated failure time regression.

    The cumulative hazard function :math:`H` is linked to the multiplier
    function :math:`g` by the relation:

    .. math::

        H(t, x) = H_0\left(\dfrac{t}{g(\beta, x)}\right) = H_0(t e^{- \beta
        \cdot x})

    where :math:`x` is a vector of covariates, :math:`\beta` is the coefficient
    vector of the effect of covariates, :math:`H_0` is the baseline cumulative
    hazard function [1]_.

    |

    Parameters
    ----------
    baseline : FittableParametricLifetimeModel
        Any lifetime model that can be fitted.
    coefficients : tuple of floats (values can be None), default is (None,)
        Coefficients values of the covariate effects.

    Attributes
    ----------
    baseline : FittableParametricLifetimeModel
        The regression baseline model (lifetime model).
    covar_effect : _CovarEffect
        The regression covariate effect.
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    coefficients
    nb_params
    params
    params_names
    plot


    References
    ----------
    .. [1] Kalbfleisch, J. D., & Prentice, R. L. (2011). The statistical
        analysis of failure time data. John Wiley & Sons.

    See Also
    --------
    regression.ProportionalHazard : proportional hazard regression
    """

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def hf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.hf(t0) / self.covar_effect.g(covar)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def chf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.chf(t0)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def ichf(self, cumulative_hazard_rate: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        return self.covar_effect.g(covar) * self.baseline.ichf(cumulative_hazard_rate)

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def dhf(self, time: AnyFloat, covar: AnyFloat) -> NumpyFloat:
        t0 = time / self.covar_effect.g(covar)
        return self.baseline.dhf(t0) / self.covar_effect.g(covar) ** 2

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def jac_hf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = _broadcast_time_covar_shapes(
            time.shape, covar.shape
        )  # (), (n,) or (m, n)
        time, covar = _broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar)  # (nb_coef, m, 1)
        t0 = time / g  # (m, n)
        # p == baseline.nb_params
        baseline_jac_hf_t0 = self.baseline.jac_hf(t0)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0)  # (m, n)
        baseline_dhf_t0 = self.baseline.dhf(t0)  # (m, n)
        jac_g = np.repeat(jac_g, baseline_hf_t0.shape[-1], axis=-1)  # (nb_coef, m, n)

        jac = np.concatenate(
            (
                -jac_g
                / g**2
                * (
                    baseline_hf_t0 + t0 * baseline_dhf_t0
                ),  # (nb_coef, m, n) necessary to concatenate
                baseline_jac_hf_t0 / g,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        return jac

    @override
    @document_args(
        base_cls=ParametricLifetimeRegression, args_docstring=_covar_docstring
    )
    def jac_chf(
        self,
        time: AnyFloat,
        covar: AnyFloat,
    ) -> NumpyFloat:
        time = np.asarray(time)  # (), (n,) or (m, n)
        covar = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        out_shape = _broadcast_time_covar_shapes(
            time.shape, covar.shape
        )  # (), (n,) or (m, n)
        time, covar = _broadcast_time_covar(time, covar)  # (m, n) and (m, nb_coef)

        g = self.covar_effect.g(covar)  # (m, 1)
        jac_g = self.covar_effect.jac_g(covar)  # (nb_coef, m, 1)
        t0 = time / g  #  (m, n)
        # p == baseline.nb_params
        baseline_jac_chf_t0 = self.baseline.jac_chf(t0)  # (p, m, n)
        baseline_hf_t0 = self.baseline.hf(t0)  #  (m, n)
        jac_g = np.repeat(
            jac_g, baseline_hf_t0.shape[-1], axis=-1
        )  # (nb_coef, m, n) necessary to concatenate

        jac = np.concatenate(
            (
                -jac_g / g * t0 * baseline_hf_t0,  #  (nb_coef, m, n)
                baseline_jac_chf_t0,  # (p, m, n)
            ),
            axis=0,
        )  # (p + nb_coef, m, n)

        jac = jac.reshape((self.nb_params,) + out_shape)
        return jac



"""Nonparametric lifetime models."""

from typing import Literal, Self, final, overload

import numpy as np
from numpy.typing import NDArray

__all__: list[str] = ["ECDF", "KaplanMeier", "NelsonAalen"]


@final
class ECDF:
    """
    Empirical Cumulative Distribution Function.
    """

    _sf: NDArray[np.void] | None
    _cdf: NDArray[np.void] | None

    def __init__(self) -> None:
        self._sf = None
        self._cdf = None

    def fit(self, time: NDArray[np.float64]) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray
            Observed lifetime values.
        """
        timeline, counts = np.unique(time, return_counts=True)
        timeline = np.insert(timeline, 0, 0)

        dtype = np.dtype(
            [("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)]
        )
        self._sf = np.empty((timeline.size,), dtype=dtype)
        self._cdf = np.empty((timeline.size,), dtype=dtype)

        self._sf["timeline"] = timeline
        self._cdf["timeline"] = timeline
        cdf = np.insert(np.cumsum(counts), 0, 0) / np.sum(counts)
        self._cdf["estimation"] = cdf
        self._sf["estimation"] = 1 - cdf
        se = np.sqrt((1 - cdf) / len(time))
        self._sf["se"] = se
        self._cdf["se"] = se
        return self

    @overload
    def sf(
        self, se: Literal[False]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def sf(
        self, se: Literal[True]
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None
    ): ...
    @overload
    def sf(
        self, se: bool
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ): ...

    def sf(
        self, se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ):
        """
        The survival functions estimated values

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._sf is None:
            return None
        if se:
            return self._sf["timeline"], self._sf["estimation"], self._sf["se"]
        return self._sf["timeline"], self._sf["estimation"]

    @overload
    def cdf(
        self, se: Literal[False]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def cdf(
        self, se: Literal[True]
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None
    ): ...
    @overload
    def cdf(
        self, se: bool
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ): ...

    def cdf(
        self, se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ):
        """
        The cumulative distribution function estimated values

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._cdf is None:
            return None
        if se:
            return self._cdf["timeline"], self._cdf["estimation"], self._cdf["se"]
        return self._cdf["timeline"], self._cdf["estimation"]

    @property
    def plot(self):
        from ._plot import PlotECDF

        return PlotECDF(self)


@final
class KaplanMeier:
    r"""Kaplan-Meier estimator.

    Compute the non-parametric Kaplan-Meier estimator (also known as the product
    limit estimator) of the survival function from lifetime data.

    Notes
    -----
    For a given time instant :math:`t` and :math:`n` total observations, this
    estimator is defined as:

    .. math::

        \hat{S}(t) = \prod_{i: t_i \leq t} \left( 1 - \frac{d_i}{n_i}\right)

    where :math:`d_i` is the number of failures until :math:`t_i` and
    :math:`n_i` is the number of assets at risk just prior to :math:`t_i`.

    The variance estimation is obtained by:

    .. math::

        \widehat{Var}[\hat{S}(t)] = \hat{S}(t)^2 \sum_{i: t_i \leq t}
        \frac{d_i}{n_i(n_i - d_i)}

    which is often referred to as Greenwood's formula.

    References
    ----------
    .. [1] Lawless, J. F. (2011). Statistical models and methods for lifetime
        data. John Wiley & Sons.

    .. [2] Kaplan, E. L., & Meier, P. (1958). Nonparametric estimation from
        incomplete observations. Journal of the American statistical
        association, 53(282), 457-481.

    """

    _sf: NDArray[np.void] | None

    def __init__(self) -> None:
        self._sf = None

    def fit(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        """

        if event is None:
            event = np.ones_like(time).astype(bool)

        if entry is None:
            entry = np.zeros_like(time)

        timeline = np.unique(time)

        n = (
            (timeline <= time.reshape(-1, 1)) * (timeline >= entry.reshape(-1, 1))
        ).sum(axis=0)

        d = ((time.reshape(-1, 1) == timeline) * event.reshape(-1, 1)).sum(axis=0)
        sf = (1 - d / n).cumprod()

        with np.errstate(divide="ignore"):
            var = (sf**2) * (d / (n * (n - d))).cumsum()
            var = np.where(n > d, var, 0)

        dtype = np.dtype(
            [("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)]
        )
        self._sf = np.empty((timeline.size + 1,), dtype=dtype)
        self._sf["timeline"] = np.insert(timeline, 0, 0)
        self._sf["estimation"] = np.insert(sf, 0, 1)
        self._sf["se"] = np.insert(np.sqrt(var), 0, 0)
        return self

    @overload
    def sf(
        self, se: Literal[False]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def sf(
        self, se: Literal[True]
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None
    ): ...
    @overload
    def sf(
        self, se: bool
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ): ...

    def sf(
        self, se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ):
        """
        The survival function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._sf is None:
            return None
        if se:
            return self._sf["timeline"], self._sf["estimation"], self._sf["se"]
        return self._sf["timeline"], self._sf["estimation"]

    @property
    def plot(self):
        from ._plot import PlotKaplanMeier

        return PlotKaplanMeier(self)


@final
class NelsonAalen:
    r"""Nelson-Aalen estimator.

    Compute the non-parametric Nelson-Aalen estimator of the cumulative hazard
    function from lifetime data.

    Notes
    -----
    For a given time instant :math:`t` and :math:`n` total observations, this
    estimator is defined as:

    .. math::

        \hat{H}(t) = \sum_{i: t_i \leq t} \frac{d_i}{n_i}

    where :math:`d_i` is the number of failures until :math:`t_i` and
    :math:`n_i` is the number of assets at risk just prior to :math:`t_i`.

    The variance estimation is obtained by:

    .. math::

        \widehat{Var}[\hat{H}(t)] = \sum_{i: t_i \leq t} \frac{d_i}{n_i^2}

    Note that the alternative survivor function estimate:

    .. math::

        \tilde{S}(t) = \exp{(-\hat{H}(t))}

    is sometimes suggested for the continuous-time case.

    References
    ----------
    .. [1] Lawless, J. F. (2011). Statistical models and methods for lifetime
        data. John Wiley & Sons.
    """

    _chf: NDArray[np.void] | None

    def __init__(self) -> None:
        self._chf = None

    def fit(
        self,
        time: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
    ) -> Self:
        """
        Compute the non-parametric estimations with respect to lifetime data.

        Parameters
        ----------
        time : ndarray
            Observed lifetime values.
        event : ndarray of boolean values (1d), default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : ndarray of float (1d), default is None
            Left truncations applied to lifetime values.
        """

        if event is None:
            event = np.ones_like(time).astype(bool)

        if entry is None:
            entry = np.zeros_like(time)

        timeline = np.unique(time)

        n = (
            (timeline <= time.reshape(-1, 1)) * (timeline >= entry.reshape(-1, 1))
        ).sum(axis=0)

        d = ((time.reshape(-1, 1) == timeline) * event.reshape(-1, 1)).sum(axis=0)

        chf = (d / n).cumsum()

        with np.errstate(divide="ignore"):
            var = (d / n**2).cumsum()

        dtype = np.dtype(
            [("timeline", np.float64), ("estimation", np.float64), ("se", np.float64)]
        )
        self._chf = np.empty((timeline.size + 1,), dtype=dtype)
        self._chf["timeline"] = np.insert(timeline, 0, 0)
        self._chf["estimation"] = np.insert(chf, 0, 0)
        self._chf["se"] = np.insert(np.sqrt(var), 0, 0)
        return self

    @overload
    def chf(
        self, se: Literal[False]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | None: ...
    @overload
    def chf(
        self, se: Literal[True]
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]] | None
    ): ...
    @overload
    def chf(
        self, se: bool
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ): ...

    def chf(
        self, se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
        | None
    ):
        """
        The cumulative hazard function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline, the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if self._chf is None:
            return None
        if se:
            return self._chf["timeline"], self._chf["estimation"], self._chf["se"]
        return self._chf["timeline"], self._chf["estimation"]

    @property
    def plot(self):
        from ._plot import PlotNelsonAalen

        return PlotNelsonAalen(self)



from typing import (
    TypeAlias,
)

import numpy as np
from numpy.typing import NDArray

__all__ = [
    "NumpyFloat",
    "NumpyBool",
    "AnyFloat",
]


# np.float64 is a subtype of float. See np.float64.mro()
AnyFloat: TypeAlias = float | NDArray[np.float64]
NumpyFloat: TypeAlias = np.float64 | NDArray[np.float64]
NumpyBool: TypeAlias = np.bool | NDArray[np.bool_]



import numpy as np
from numpy.typing import NDArray


def select_from_struct(
    struct_array,
    sample_id=None,
    asset_id=None,
):
    """
    Method used for dev and tests
    """
    mask: NDArray[np.bool_] = np.ones_like(struct_array, dtype=np.bool_)
    if sample_id is not None:
        mask = mask & np.isin(struct_array["sample_id"], sample_id)
    if asset_id is not None:
        mask = mask & np.isin(struct_array["asset_id"], asset_id)
    return struct_array[mask].copy()



"""Base classes for all parametric lifetime models."""

from __future__ import annotations

import copy
import functools
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    ParamSpec,
    Self,
    TypeVar,
    TypeVarTuple,
    final,
    overload,
)

import numpy as np
import numpydoc.docscrape as docscrape  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import NDArray
from optype.numpy import Array1D, ToFloat
from scipy.optimize import newton
from typing_extensions import override

from relife.base import (
    FittingResults,
    FrozenParametricModel,
    MaximumLikelihoodOptimizer,
    OptimizerConfig,
    ParametricModel,
)
from relife.typing import (
    AnyFloat,
    NumpyBool,
    NumpyFloat,
    Seed,
)
from relife.utils import reshape_1d_arg
from relife.utils.quadrature import legendre_quadrature, unweighted_laguerre_quadrature

__all__ = [
    "ParametricLifetimeModel",
    "FittableParametricLifetimeModel",
    "LifetimeData",
    "LifetimeLikelihood",
]

Ts = TypeVarTuple("Ts")


class ParametricLifetimeModel(ParametricModel, ABC, Generic[*Ts]):
    r"""Base class for parametric lifetime models in ReLife.

    This class is a blueprint for implementing parametric lifetime models. The
    interface is generic and can define a variadic set of arguments. It expects
    implementation of the hazard function (`hf`), the cumulative hazard
    function (`chf`), the probability density function (`pdf`) and the survival
    function (`sf`). Other functions are implemented by default but can be
    overridden by the derived classes.

    Note:
        The abstract methods also provides a default implementation. One may
        not have to implement `hf`, `chf`, `pdf` and `sf` and just call
        `super()` to access the base implementation.

    Methods:
        hf: Abstract method to compute the hazard function.
        chf: Abstract method to compute the cumulative hazard function.
        sf: Abstract method to compute the survival function.
        pdf: Abstract method to compute the probability density function.
    """

    @abstractmethod
    def sf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            sf values at each given time(s).
        """
        if hasattr(self, "chf"):
            return np.exp(
                -self.chf(
                    time,
                    *args,
                )
            )
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return self.pdf(time, *args) / self.hf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete sf function
                """
            )

    @abstractmethod
    def hf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            hf values at each given time(s).
        """
        if hasattr(self, "pdf") and hasattr(self, "sf"):
            return self.pdf(time, *args) / self.sf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                    {class_name} must implement concrete hf function
                    """
            )

    @abstractmethod
    def chf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            chf values at each given time(s).
        """
        if hasattr(self, "sf"):
            return -np.log(self.sf(time, *args))
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return -np.log(self.pdf(time, *args) / self.hf(time, *args))
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete chf or at least concrete hf function
                """
            )

    @abstractmethod
    def pdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            pdf values at each given time(s).
        """
        try:
            return self.sf(time, *args) * self.hf(time, *args)
        except NotImplementedError as err:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
            {class_name} must implement pdf or the above functions
            """
            ) from err

    def cdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            cdf values at each given time(s).
        """
        return 1 - self.sf(time, *args)

    def ppf(self, probability: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The percent point function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            ppf values at each given probability value(s).
        """
        probability = np.asarray(probability)
        return self.isf(1 - probability, *args)

    def median(self, *args: *Ts) -> NumpyFloat:
        """
        The median.

        Parameters
        ----------
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
        """
        return self.ppf(0.5, *args)

    def isf(self, probability: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The inverse survival function.

        Parameters
        ----------
        probability : float or np.ndarray
            Probability value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            isf values at each given probability value(s).
        """

        def func(x: NDArray[np.float64]) -> NumpyFloat:
            return self.sf(x, *args) - probability

        # no idea on how to type func
        return newton(  # pyright: ignore[reportCallIssue, reportUnknownVariableType]
            func,  # pyright: ignore[reportArgumentType]
            x0=np.zeros_like(probability),
            args=args,
        )

    def ichf(self, cumulative_hazard_rate: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        Inverse cumulative hazard function.

        Parameters
        ----------
        cumulative_hazard_rate : float or np.ndarray
            Cumulative hazard rate value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            ichf values at each given cumulative hazard rate(s).
        """

        def func(x: NDArray[np.float64]) -> NumpyFloat:
            return self.chf(x, *args) - cumulative_hazard_rate

        # no idea on how to type func
        return newton(  # pyright: ignore[reportCallIssue, reportUnknownVariableType]
            func,  # pyright: ignore[reportArgumentType]
            x0=np.zeros_like(cumulative_hazard_rate),
        )

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    def rvs(
        self,
        size: int | tuple[int, int],
        *args: *Ts,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        """
        Random variable sampling.

        Parameters
        ----------
        size : int or tuple (m, n) of int
            Size of the generated sample.
        *args
            Any additonal args.
        return_event : bool, default is False
            If True, returns event indicators along with the sample time
            values.
        return_entry : bool, default is False
            If True, returns corresponding entry values of the sample time
            values.
        seed : optional int, np.random.BitGenerator, np.random.Generator, np.random.RandomState, default is None
            If int or BitGenerator, seed for random number generator. If
            np.random.RandomState or np.random.Generator, use as given.

        Returns
        -------
        out : float, ndarray or tuple of float or ndarray
            The sample values. If either `return_event` or `return_entry` is
            True, returns a tuple containing the time values followed by event
            values, entry values or both.
        """
        rng = np.random.default_rng(seed)
        probability = rng.uniform(size=size)
        if size == 1:
            probability = np.squeeze(probability)
        time = self.isf(probability, *args)
        event = (
            np.ones_like(time, dtype=np.bool_)
            if isinstance(time, np.ndarray)
            else np.bool_(True)
        )
        entry = (
            np.zeros_like(time, dtype=np.float64)
            if isinstance(time, np.ndarray)
            else np.float64(0)
        )
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            return time, event
        elif not return_event and return_entry:
            return time, entry
        else:
            return time, event, entry

    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        *args: *Ts,
        deg: int = 10,
    ) -> NumpyFloat:
        """
        Lebesgue-Stieltjes integration.

        Parameters
        ----------
        func : callable (in : 1 ndarray , out : 1 ndarray)
            The callable must have only one ndarray object as argument and one
            ndarray object as output.
        a : ndarray (maximum number of dimension is 2)
            Lower bound(s) of integration.
        b : ndarray (maximum number of dimension is 2)
            Upper bound(s) of integration. If lower bound(s) is infinite, use
            np.inf as value.
        *args
            Any additonal args.
        deg : int, default 10
            Degree of the polynomials interpolation.

        Returns
        -------
        out : np.ndarray
            Lebesgue-Stieltjes integral of func from `a` to `b`.
        """

        def integrand(x: NDArray[np.float64]) -> NDArray[np.float64]:
            #  x.shape == (deg,), (deg, n) or (deg, m, n), ie points of quadratures
            # fx : (d_1, ..., d_i, deg), (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            x = np.asarray(x)
            fx = func(x)

            try:
                _ = np.broadcast_shapes(fx.shape[-len(x.shape) :], x.shape)
            except ValueError:
                raise ValueError(
                    """
                    func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
                    Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
                    """
                )
            if (
                x.ndim == 3
            ):  # reshape because model.pdf is tested only for input ndim <= 2
                x_shape: tuple[int, int, int] = x.shape
                xdeg, m, n = x_shape
                x = np.rollaxis(x, 1).reshape(
                    m, -1
                )  # (m, deg*n), roll on m because axis 0 must align with m of args
                pdf = self.pdf(x, *args)  # (m, deg*n)
                pdf = np.rollaxis(pdf.reshape(m, xdeg, n), 1, 0)  #  (deg, m, n)
            else:  # ndim == 1 | 2
                # reshape to (1, deg*n) or (1, deg), ie place 1 on axis 0 to allow broadcasting with m of args
                pdf = self.pdf(x.reshape(1, -1), *args)  # (1, deg*n) or (1, deg)
                pdf = pdf.reshape(x.shape)  # (deg, n) or (deg,)

            # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            return fx * pdf

        arr_a, arr_b = np.broadcast_arrays(a, b)  # (), (n,) or (m, n)
        if np.any(arr_a > arr_b):
            raise ValueError("Bound values a must be lower than values of b")

        bound_b = self.isf(
            1e-4, *args
        )  #  () or (m, 1), if (m, 1) then arr_b.shape == (m, 1) or (m, n)
        broadcasted_arrs = np.broadcast_arrays(arr_a, arr_b, bound_b)
        arr_a = broadcasted_arrs[
            0
        ].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        arr_b = broadcasted_arrs[
            1
        ].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        bound_b = broadcasted_arrs[
            2
        ].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        is_inf = np.isinf(arr_b)  # () or (n,) or (m, n)
        arr_b = np.where(is_inf, bound_b, arr_b)
        integration = legendre_quadrature(
            integrand, arr_a, arr_b, deg=deg
        )  #  (d_1, ..., d_i), (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)
        is_inf, _ = np.broadcast_arrays(is_inf, integration)
        return np.where(
            is_inf,
            integration + unweighted_laguerre_quadrature(integrand, arr_b, deg=deg),
            integration,
        )

    def moment(self, n: int, *args: *Ts) -> NumpyFloat:
        """
        n-th order moment.

        Parameters
        ----------
        n : int
            order of the moment, at least 1.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64
        """
        if n < 1:
            raise ValueError("order of the moment must be at least 1")

        def func(x: NDArray[np.float64]) -> NDArray[np.float64]:
            return np.power(x, n)

        return self.ls_integrate(
            func,
            0.0,
            np.inf,
            *args,
            deg=100,
        )  #  high degree of polynome to ensure high precision

    def mean(self, *args: *Ts) -> NumpyFloat:
        """
        The mean of the distribution.

        Parameters
        ----------
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
        """
        return self.moment(1, *args)

    def var(self, *args: *Ts) -> NumpyFloat:
        """
        The variance of the distribution.

        Parameters
        ----------
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
        """
        return self.moment(2, *args) - self.moment(1, *args) ** 2

    def mrl(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The mean residual life function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            Function values at each given time(s).
        """

        sf = self.sf(time, *args)

        def func(x: NDArray[np.float64]) -> NDArray[np.float64]:
            return np.asarray(x) - time

        ls = self.ls_integrate(func, time, np.inf, *args)
        if sf.ndim < 2:  # 2d to 1d or 0d
            ls = np.squeeze(ls)
        return ls / sf

    @property
    def plot(self):
        """Provides access to plotting functionnalities"""
        from ._plot import PlotParametricLifetimeModel

        return PlotParametricLifetimeModel(self)


class FrozenParametricLifetimeModel(
    FrozenParametricModel[ParametricLifetimeModel[*Ts], *Ts]
):
    _args: tuple[*Ts]
    _unfrozen_model: ParametricLifetimeModel[*Ts]

    def __init__(self, model: ParametricLifetimeModel[*Ts], *args: *Ts) -> None:
        super().__init__(model, *args)

    def sf(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.sf(time, *self._args)

    def hf(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.hf(time, *self._args)

    def chf(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.chf(time, *self._args)

    def pdf(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.pdf(time, *self._args)

    def cdf(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.cdf(time, *self._args)

    def ppf(self, probability: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.ppf(probability, *self._args)

    def median(self) -> NumpyFloat:
        return self._unfrozen_model.median(*self._args)

    def isf(self, probability: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.isf(probability, *self._args)

    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.ichf(cumulative_hazard_rate, *self._args)

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    def rvs(
        self,
        size: int | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        return self._unfrozen_model.rvs(
            size,
            *self._args,
            return_event=return_event,
            return_entry=return_entry,
            seed=seed,
        )

    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: NumpyFloat,
        b: NumpyFloat,
        *,
        deg: int = 10,
    ) -> NumpyFloat:
        return self._unfrozen_model.ls_integrate(func, a, b, *self._args, deg=deg)

    def moment(self, n: int) -> NumpyFloat:
        return self._unfrozen_model.moment(n, *self._args)

    def mean(self) -> NumpyFloat:
        return self._unfrozen_model.mean(*self._args)

    def var(self) -> NumpyFloat:
        return self._unfrozen_model.var(*self._args)

    def mrl(self, time: AnyFloat) -> NumpyFloat:
        return self._unfrozen_model.mrl(time, *self._args)


@overload
def is_lifetime_model(model: FrozenParametricLifetimeModel[*Ts]) -> Literal[True]: ...
@overload
def is_lifetime_model(
    model: ParametricLifetimeModel[*Ts]
    | FrozenParametricLifetimeModel[ParametricModel, *Ts],
) -> Literal[True]: ...
@overload
def is_lifetime_model(
    model: Any
    | ParametricLifetimeModel[*Ts]
    | FrozenParametricLifetimeModel[ParametricModel, *Ts],
) -> bool: ...
def is_lifetime_model(
    model: Any
    | ParametricLifetimeModel[*Ts]
    | FrozenParametricLifetimeModel[ParametricModel, *Ts],
) -> bool:
    """
    Checks if model is a lifetime model.
    """
    return isinstance(model, (ParametricLifetimeModel, FrozenParametricLifetimeModel))


P = ParamSpec("P")
T = TypeVar("T")


def document_args(
    *,
    base_cls: type,
    args_docstring: list[docscrape.Parameter],
    returns: list[docscrape.Parameter] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def decorator_extend_docstring(
        method: Callable[P, T],
    ) -> Callable[P, T]:
        base_doc = getattr(base_cls, method.__name__).__doc__
        numpy_docstring = docscrape.NumpyDocString(base_doc)
        new_parameters_docstring: list[docscrape.Parameter] = []
        for param in numpy_docstring["Parameters"]:  # pyright: ignore[reportUnknownVariableType]
            if param.name != "*args":  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                new_parameters_docstring.append(param)  # pyright: ignore[reportArgumentType]
            else:
                new_parameters_docstring += args_docstring
        numpy_docstring["Parameters"] = new_parameters_docstring
        if returns is not None:
            numpy_docstring["Returns"] = returns
        method.__doc__ = str(numpy_docstring)

        @functools.wraps(method)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return method(*args, **kwargs)

        return wrapper

    return decorator_extend_docstring


M = TypeVar(
    "M",
    bound="FittableParametricLifetimeModel[*tuple[Any, ...]]",
)


class FittableParametricLifetimeModel(ParametricLifetimeModel[*Ts], ABC):
    fitting_results: FittingResults | None

    def __init__(self, **kwparams: float | None):
        super().__init__(**kwparams)
        self.fitting_results = None

    @abstractmethod
    def jac_hf(
        self,
        time: AnyFloat,
        *args: *Ts,
    ) -> NumpyFloat:
        """
        The jacobian of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            The derivatives with respect to each parameter. If the result is
            an `np.ndarray`, the first dimension holds the number of parameters.
        """

    @abstractmethod
    def jac_chf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The jacobian of the cumulative hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            The derivatives with respect to each parameter. If the result is
            an `np.ndarray`, the first dimension holds the number of parameters.
        """

    @abstractmethod
    def jac_sf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The jacobian of the survival function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            The derivatives with respect to each parameter. If the result is
            an `np.ndarray`, the first dimension holds the number of parameters.
        """

    def jac_cdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The jacobian of the cumulative density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            The derivatives with respect to each parameter. If the result is
            an `np.ndarray`, the first dimension holds the number of parameters.
        """
        return -self.jac_sf(time, *args)

    @abstractmethod
    def jac_pdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The jacobian of the probability density function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            The derivatives with respect to each parameter. If the result is
            an `np.ndarray`, the first dimension holds the number of parameters.

        """

    @abstractmethod
    def dhf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The derivate of the hazard function.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are `()`, `(n,)` or `(m, n)`.
        *args
            Any additonal args.

        Returns
        -------
        out : np.float64 or np.ndarray
            Function values at each given time(s).
        """

    @abstractmethod
    def init_likelihood(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> LifetimeLikelihood[M]:
        r"""
        Initialize the lifetime likelihood used to fit the parameters.

        `fit` method is the preferred way to fit model parameters. However,
        users can also interact with the likelihood returned by
        `init_likelihood` to study the optimization process.

        This method implementation is usally composed of 3 steps:
            1. Initialize an object to preprocess and encapsulate observation values.
            2. Create a `OptimizerConfig` config instance depending on the model needs.
            3. Instanciate and return a LifetimeLikelihood.

        `init_likelihood` is separated from `fit` in order to reuse existing
        likelihood parametrization in case of model composition. Any parameters
        initialization needed by the likelihood optimizer (e.g. `x0` or
        `bounds` as required in step 2.) are left to specific functions
        alongside concrete model implementations. These functions are invoked
        within `init_likelihood`.

        Parameters
        ----------
        time : 1d array
            Observed lifetime values.
        args : any ndarray or tuple of ndarray, default is None
            Additional arguments required by the model (e.g. covar).
        event : 1d array of bool, default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : 1d array, default is None
            Left truncations applied to lifetime values.
        **kwargs
            Extra arguments to control the parameters optimization. It can be:

                - those used by `scipy.optimize.minimize
                  <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
                  to search for the paremeters that minimize the negative
                  log-likelihood.
                - `covariance_method` to control the method used to estimate
                  parameters covariance. Values can be `"cs"`, `"2point"`,
                  `"exact"` or `False`. To skip parameters covariance
                  estimation, set `covariance_method` to `False`, otherwise the
                  default method associated to the model will be used. If
                  `covariance_method` is `"exact"` the `hess` must be passed
                  too.

        Returns
        -------
        out : LifetimeLikelihood instance
        """

    def fit(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> Self:
        """
        Estimation of parameters from lifetime data.

        Parameters
        ----------
        time : 1d array
            Observed lifetime values.
        args : any ndarray or tuple of ndarray, default is None
            Additional arguments required by the model (e.g. covar).
        event : 1d array of bool, default is None
            Boolean indicators tagging lifetime values as right censored or complete.
        entry : 1d array, default is None
            Left truncations applied to lifetime values.
        **kwargs
            Extra arguments used by `scipy.optimize.minimize
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            to search for the paremeters that minimize the negative
            log-likelihood. `covariance_method` can also be passed to control
            the method used to estimate parameters covariance. Values can be
            `"cs"`, `"2point"`, `"exact"` or `False`. To skip parameters
            covariance estimation, set `covariance_method` to `False`,
            otherwise the default method associated to the model will be used.
            If `covariance_method` is `"exact"` the `hess` must be passed too.

        Returns
        -------
        out : the object instance
            The estimated parameters are setted inplace.
        """
        optimizer: LifetimeLikelihood[Self] = self.init_likelihood(
            time, args, event, entry, **kwargs
        )
        assert id(optimizer.model) != id(self)
        self.fitting_results = optimizer.optimize()
        self.params: NDArray[np.float64] = self.fitting_results.optimal_params

        return self


@dataclass
class LifetimeData:
    time: NDArray[np.float64]  # 1d array or 2d
    args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    event: NDArray[np.bool_] | None = None
    entry: NDArray[np.float64] | None = None

    nb_observations: int = field(init=False)
    complete_time: NDArray[np.float64] = field(init=False, repr=False)
    censored_time: NDArray[np.float64] = field(init=False, repr=False)
    left_truncations: NDArray[np.float64] = field(init=False, repr=False)
    complete_time_args: tuple[NDArray[Any], ...] = field(init=False, repr=False)
    censored_time_args: tuple[NDArray[Any], ...] = field(init=False, repr=False)
    left_truncations_args: tuple[NDArray[Any], ...] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        time = reshape_1d_arg(self.time)
        if time.shape[-1] == 2 and self.event is not None:
            raise ValueError("If time is given as intervals, event must be None")
        event = None
        if time.shape[-1] == 1:
            event = (
                reshape_1d_arg(self.event)
                if self.event is not None
                else np.ones_like(time, dtype=np.bool_)
            )
        entry = (
            reshape_1d_arg(self.entry)
            if self.entry is not None
            else np.zeros(len(time), dtype=np.float64)
        )
        if np.any(time <= entry):
            raise ValueError("All time values must be greater than entry values")
        if isinstance(self.args, tuple):
            args = tuple((reshape_1d_arg(arg) for arg in self.args))
        elif isinstance(self.args, np.ndarray):
            args = (reshape_1d_arg(self.args),)
        else:
            args = ()
        sizes = [len(x) for x in (time, event, entry, *args) if x is not None]
        if len(set(sizes)) != 1:
            raise ValueError(
                f"""
                All lifetime data must have the same number of values. Fields
                length are different. Got {tuple(sizes)}
                """
            )
        non_zero_entry = np.flatnonzero(entry)
        if event is not None:
            non_zero_event = np.flatnonzero(event)
            zero_event = np.flatnonzero(event == 0)
            self.nb_observations = time.size
            self.complete_time = time[non_zero_event]
            self.censored_time = time[zero_event]
            self.left_truncations = entry[non_zero_entry]
            self.complete_time_args = tuple(arg[non_zero_event] for arg in args)
            self.censored_time_args = tuple(arg[zero_event] for arg in args)
            self.left_truncations_args = tuple(arg[non_zero_entry] for arg in args)
        else:
            complete_time_index = np.flatnonzero(time[:, 0] == time[:, 1])
            non_complete_time_index = np.flatnonzero(time[:, 0] != time[:, 1])
            self.nb_observations = time.size
            self.complete_time = time[:, 1][complete_time_index]
            self.censored_time = time[non_complete_time_index]
            self.left_truncations = entry[non_zero_entry]
            self.complete_time_args = tuple(arg[complete_time_index] for arg in args)
            self.censored_time_args = tuple(
                arg[non_complete_time_index] for arg in args
            )
            self.left_truncations_args = tuple(arg[non_zero_entry] for arg in args)


@final
class LifetimeLikelihood(MaximumLikelihoodOptimizer[M, LifetimeData]):
    """
    Maximum likelihood estimator from lifetime data.

    Parameters
    ----------
    model : generic FittableParametricLifetimeModel
        Every model parameters must be initialized before passing it to the
        likelihood.
    data : LifetimeData
        An object that encapsulate and preprocess lifetime observations and
        truncations.
    config : OptimizerConfig
        An object that groups configurations used by the optimizer.

    Attributes
    ----------
    model: FittableParametricLifetimeModel
        A copy of the original model.
    data : LifetimeData
        An object that encapsulate and preprocess lifetime observations and
        truncations.
    config : OptimizerConfig
        An object that groups configurations used by the optimizer.
    """

    model: M
    data: LifetimeData

    def __init__(
        self,
        model: M,
        data: LifetimeData,
        config: OptimizerConfig,
    ):
        self.model = copy.deepcopy(model)
        self.data = data
        self.config = config
        if "jac" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["jac"] = self.jac_negative_log

    @property
    @override
    def nb_observations(self) -> int:
        return self.data.nb_observations

    @override
    def negative_log(self, params: Array1D[np.float64]) -> ToFloat:
        self.model.params = params
        return (
            _complete_time_contrib(self.model, self.data)
            + _censored_time_contrib(self.model, self.data)
            + _left_truncations_contrib(self.model, self.data)
        )

    def jac_negative_log(self, params: Array1D[np.float64]) -> Array1D[np.float64]:
        """
        Jacobian of the negative log likelihood.

        The jacobian is computed with respect to parameters.

        Parameters
        ----------
        model : parametric model
            A parametrized model with appropriate parameters values.

        Returns
        -------
        out : ndarray
        """
        self.model.params = params
        return (
            _jac_complete_time_contrib(self.model, self.data)
            + _jac_censored_time_contrib(self.model, self.data)
            + _jac_left_truncations_contrib(self.model, self.data)
        )


def _complete_time_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> float:
    if data.complete_time.size == 0.0:
        return 0.0
    return -np.sum(np.log(model.pdf(data.complete_time, *data.complete_time_args)))


def _jac_complete_time_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> NDArray[np.float64]:
    if data.complete_time.size == 0:
        return np.zeros_like(model.params)
    jac = -model.jac_pdf(data.complete_time, *data.complete_time_args) / model.pdf(
        data.complete_time, *data.complete_time_args
    )

    return np.sum(jac, axis=(1, 2))


def _censored_time_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> float:
    if data.censored_time.size == 0:
        return 0.0
    if data.censored_time.shape[-1] > 1:
        # interval censored time
        return np.sum(
            -np.log(
                10**-10
                + model.cdf(data.censored_time[:, 1], *data.censored_time_args)
                - model.cdf(data.censored_time[:, 0], *data.censored_time_args)
            ),
        )
    else:
        # right censored time
        return np.sum(model.chf(data.censored_time, *data.censored_time_args))


def _jac_censored_time_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> NDArray[np.float64]:
    if data.censored_time.size == 0:
        return np.zeros_like(model.params)
    if data.censored_time.shape[-1] > 1:
        # interval censored time
        jac_interval_censored = (
            model.jac_sf(data.censored_time[:, 1], *data.censored_time_args)
            - model.jac_sf(data.censored_time[:, 0], *data.censored_time_args)
        ) / (
            10**-10
            + model.cdf(data.censored_time[:, 1], *data.censored_time_args)
            - model.cdf(data.censored_time[:, 0], *data.censored_time_args)
        )

        return np.sum(jac_interval_censored, axis=(1, 2))
    else:
        # right censored time
        return np.sum(
            model.jac_chf(data.censored_time, *data.censored_time_args),
            axis=(1, 2),
        )


def _left_truncations_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> float:
    if data.left_truncations.size == 0.0:
        return 0.0
    return -np.sum(model.chf(data.left_truncations, *data.left_truncations_args))


def _jac_left_truncations_contrib(
    model: FittableParametricLifetimeModel[*tuple[Any, ...]],
    data: LifetimeData,
) -> NDArray[np.float64]:
    if data.left_truncations.size == 0.0:
        return np.zeros_like(model.params)
    jac = -model.jac_chf(data.left_truncations, *data.left_truncations_args)
    return np.sum(jac, axis=(1, 2))



from typing import Any, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

from relife.typing import NumpyFloat

__all__ = [
    "reshape_1d_arg",
    "flatten_if_possible",
    "get_args_nb_assets",
]


E = TypeVar("E", bound=np.generic, covariant=True)


@overload
def reshape_1d_arg(arg: int) -> np.float64: ...
@overload
def reshape_1d_arg(arg: float) -> np.float64: ...
@overload
def reshape_1d_arg(arg: bool | np.bool_) -> np.bool: ...
@overload
def reshape_1d_arg(arg: NDArray[E]) -> NDArray[E]: ...
def reshape_1d_arg(
    arg: int | float | bool | np.bool | NDArray[E],
) -> np.float64 | np.bool | NDArray[E]:
    """
    Reshapes ReLife arguments that are expected to be 0d or 1d.

    Parameters
    ----------
    arg : float, 0d or 1d array

    Returns
    -------
    np.float64 or (m, 1) shaped array
        Reshaped array used to ensure broadcasting compatibility in computations.
    """
    if isinstance(arg, (int, float)):  # np.float64 is float
        return np.float64(arg)
    if isinstance(arg, (bool, np.bool)):
        return np.bool(arg)
    if arg.ndim == 1:
        arg = np.atleast_1d(arg).reshape(-1, 1)
    if arg.ndim > 2:
        raise ValueError("arg can't be more than 2d")
    return arg


def flatten_if_possible(value: NumpyFloat) -> NumpyFloat:
    """
    Flatten array-like object when possible.

    Parameters
    ----------
    value : np.ndarray

    Returns
    -------
    np.ndarray
        Flattened array.
    """
    if value.ndim != 0:
        return value.flatten()
    return value


def get_args_nb_assets(*args: NDArray[Any]) -> int:
    """
    Gets the number of assets encoded in args.
    """
    if not bool(args):
        return 1
    reshaped_args = tuple((np.atleast_2d(arg) for arg in args))
    try:
        broadcast_shape = np.broadcast_shapes(*(ary.shape for ary in reshaped_args))
    except ValueError:
        raise ValueError("args have incompatible shapes")
    if len(broadcast_shape) == 0:
        return 1
    return broadcast_shape[0]



# pyright: basic

import numpy as np

from relife.stochastic_process import NonHomogeneousPoissonProcess
from relife.stochastic_process._sample._iterables import (
    NonHomogeneousPoissonProcessIterable,
)
from relife.stochastic_process._sample.tests.utils import select_from_struct


def test_basic_sampling(distribution):
    nhpp = NonHomogeneousPoissonProcess(distribution)
    nb_samples = 10
    t0 = distribution.ppf(0.3)
    tf = distribution.ppf(0.95)

    iterable = NonHomogeneousPoissonProcessIterable(nhpp, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # Check NHPP property for each sample
    for i in range(nb_samples):
        select_sample = select_from_struct(struct_array, sample_id=i)
        np.testing.assert_equal(select_sample["time"][:-1], select_sample["entry"][1:])


def test_age_replacement_sampling(frozen_ar_distribution):
    nhpp = NonHomogeneousPoissonProcess(frozen_ar_distribution)
    ar = frozen_ar_distribution.args[0]
    nb_assets = 1 if isinstance(ar, float) else ar.shape[0]
    nb_samples = 10
    t0 = frozen_ar_distribution.ppf(0.3)
    tf = frozen_ar_distribution.ppf(0.95)

    iterable = NonHomogeneousPoissonProcessIterable(nhpp, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # Check that all times are less than ar for each asset
    for i in range(nb_assets):
        ar_asset = ar if isinstance(ar, float) else ar[i]
        select_asset = select_from_struct(struct_array, asset_id=i)
        np.testing.assert_array_less(select_asset["time"], ar_asset + 1e-5)


def test_regression_sampling(frozen_regression):
    nhpp = NonHomogeneousPoissonProcess(frozen_regression)
    nb_assets = frozen_regression.args[0].shape[0]
    t0 = frozen_regression.ppf(0.25).min()
    tf = frozen_regression.ppf(0.95).min()
    nb_samples = 10

    iterable = NonHomogeneousPoissonProcessIterable(nhpp, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # Check NHPP proprerty for each sample and each asset
    for i in range(nb_assets):
        for j in range(nb_samples):
            select_sample = select_from_struct(struct_array, asset_id=i, sample_id=j)
            np.testing.assert_equal(
                select_sample["time"][:-1], select_sample["entry"][1:]
            )


def test_age_replacement_regression_sampling(frozen_ar_regression):
    nhpp = NonHomogeneousPoissonProcess(frozen_ar_regression)
    t0 = frozen_ar_regression.ppf(0.25).min()
    tf = 10 * frozen_ar_regression.ppf(0.75).max()
    nb_samples = 10

    iterable = NonHomogeneousPoissonProcessIterable(nhpp, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # check all times are bounded by the age of replacement
    # add a small constant for numerical approximations
    for i in range(frozen_ar_regression.args[0].shape[0]):
        times = select_from_struct(struct_array, asset_id=i)["time"]
        np.testing.assert_array_less(
            times, frozen_ar_regression.args[0][i].item() + 1e-5
        )
